//
//  Converter.m
//

#import "Converter.h"
#import <CoreAudio/CoreAudioTypes.h>
#import "Functions.h"
#import "InputFile.h"

typedef struct {
	uint64_t bytes;
	uint64_t nanoseconds;
} speedStatistics_t;

@interface Converter ()

@property (assign) unsigned long inputCount;
@property (assign) unsigned long inputIndex;
@property (assign) uint64_t totalInputSize;
@property (assign) uint64_t definitiveProgress;
@property (assign) BOOL isInterrupted;
@property (assign) uint64_t taskStartTime;
@property (assign) uint64_t startTime;
@property (assign) speedStatistics_t speedStats;
@property (assign) double speed;
@property (assign) uint64_t inputSize;

@property (strong) NSMutableArray *commonArgs;
@property (strong) NSMutableArray *inputFiles;
@property (strong) NSString *outputDirectoryPath;
@property (strong) NSMutableArray *faultyInputNames;
@property (strong) NSTimer *timer;
@property (strong) NSTask *task;
@property (strong) NSString *outputPath;

- (void)initCommonArgs;
- (void)initInputFiles;
- (void)initOutputDirectoryPath;
- (void)initFaultyInputNames;
- (void)initSpeed;
- (void)initTimer;
- (void)startTimer;
- (void)reactToTimer:(NSTimer *)timer;
- (void)updateSpeedStats;
- (void)updateSpeed;
- (void)updateProgress;
- (void)cleanupFileSystem;
- (void)terminate;
- (void)flagAbortedFileNames;
- (void)flagTheFileAtIndex:(unsigned long)index;
- (void)feedback;
- (void)processNextFile;
- (void)initTaskStartTime;
- (void)initOutputPath;
- (void)initTask;
- (NSString *)finalOutputPath;
- (void)readSpeed;
- (void)writeSpeed;

@end

@implementation Converter

//--------------------------------------
#pragma mark property implementations

synth(options);
synth(inputURLs);
synth(outputDirectoryURL);
synth(feedbacker);

synth(inputCount);
synth(inputIndex);
synth(totalInputSize);
synth(definitiveProgress);
synth(isInterrupted);
synth(taskStartTime);
synth(startTime);
synth(speedStats);
synth(speed);
synth(inputSize);

synth(commonArgs);
synth(inputFiles);
synth(outputDirectoryPath);
synth(faultyInputNames);
synth(timer);
synth(task);
synth(outputPath);

//--------------------------------------
#pragma mark public methods

- (void)start
{
	[self initCommonArgs];
	[self initInputFiles];
	[self initOutputDirectoryPath];
	[self initFaultyInputNames];
	[self readSpeed];
	[self initSpeed];
	[_feedbacker feedbackTotalInputSize:_totalInputSize];
	[self initTimer];
	[self startTimer];
	[self processNextFile];
}

- (void)stop
{
	[self setIsInterrupted:YES];
	[_task interrupt];
	[_timer fire];
}

//--------------------------------------
#pragma mark private methods

#define maxNumberOfArguments 18
#define maxSampleRate 1000000
#define maxBitRate (1000 * 1000000)

- (void)initCommonArgs
{
	NSMutableArray *args = [NSMutableArray arrayWithCapacity:maxNumberOfArguments];
	dataFormat_t *dataFormat = [_options dataFormat];
	
	fileFormat_t *fileFormat = [_options fileFormat];
	[args addObject:@"--file"];
	[args addObject:fileFormat->code];
	
	NSString *dataFormatString = nil;
	unsigned sampleRate = [_options sampleRate];
	if (sampleRate > 0 && sampleRate <= maxSampleRate)
		dataFormatString = [dataFormat->code stringByAppendingFormat:@"@%u", sampleRate];
	else
		dataFormatString = dataFormat->code;
	[args addObject:@"--data"];
	[args addObject:dataFormatString];
	
	channelLayout_t *channelLayout = [_options channelLayout];
	if (dataFormat->variableLayout && channelLayout != &channelLayouts[0]) {
		unsigned layoutTag = channelLayout->tag;
		if (layoutTag == kAudioChannelLayoutTag_DiscreteInOrder)
			layoutTag |= [_options numberOfChannels];
		NSString *layoutString = [NSString stringWithFormat:@"%u", layoutTag];
		[args addObject:@"--userproperty"];
		[args addObject:@"ocl "];
		[args addObject:layoutString];
	}
	
	if (dataFormat->variableBitRate) {
		unsigned bitRate = 1000 * [_options bitRate]; // mapping kbps to bps
		if (bitRate > 0 && bitRate <= maxBitRate) {
			NSString *bitRateString = [NSString stringWithFormat:@"%u", bitRate];
			[args addObject:@"--bitrate"];
			[args addObject:bitRateString];
		}
	}
	
	if (dataFormat->variableQuality) {
		unsigned codecQuality = MIN(32 * [_options codecQuality], 127U); // mapping 0..4 to 0..127
		NSString *codecQualityString = [NSString stringWithFormat:@"%u", codecQuality];
		[args addObject:@"--quality"];
		[args addObject:codecQualityString];
	}
	
	unsigned srcQuality = MIN(32 * [_options srcQuality], 127U); // mapping 0..4 to 0..127
	NSString *srcQualityString = [NSString stringWithFormat:@"%u", srcQuality];
	[args addObject:@"--src-quality"];
	[args addObject:srcQualityString];
	
	NSString *srcComplexityCodes[] = { @"line", @"norm", @"bats" };
	
	NSString *srcComplexityString = srcComplexityCodes[[_options srcComplexity]];
	[args addObject:@"--src-complexity"];
	[args addObject:srcComplexityString];
	
	if (dataFormat->variableBitRate) {
		NSString *strategyString = [NSString stringWithFormat:@"%u", [_options strategy]];
		[args addObject:@"--strategy"];
		[args addObject:strategyString];
	}
	
	if (dataFormat->worksWithFillerOption && ![_options fillerFlag])
		[args addObject:@"--no-filler"];
	
	[self setCommonArgs:args];
}

- (void)initInputFiles
{
	unsigned long count = [_inputURLs count];
	NSMutableArray *filesArray = [NSMutableArray arrayWithCapacity:count];
	unsigned sum = 0;
	for (NSURL *url in _inputURLs) {
		InputFile *file = [InputFile newInputFileWithURL:url];
		[filesArray addObject:file];
		sum += [file size];
	}
	[self setInputCount:count];
	[self setTotalInputSize:sum];
	[self setInputFiles:filesArray];
}

- (void)initOutputDirectoryPath
{
	NSString *path = [_outputDirectoryURL path];
	[self setOutputDirectoryPath:path];
}

- (void)initFaultyInputNames
{
	NSMutableArray *names = [NSMutableArray arrayWithCapacity:_inputCount];
	[self setFaultyInputNames:names];
}

- (void)initSpeed
{
	dataFormat_t *dataFormat = [_options dataFormat];
	[self setSpeed:dataFormat->conversionSpeed];
}

#define timerFrequency 30
#define timerPeriod (1.0 / timerFrequency)

- (void)initTimer
{
	NSTimer *timer = [NSTimer timerWithTimeInterval:timerPeriod
											 target:self
										   selector:@selector(reactToTimer:)
										   userInfo:nil
											repeats:YES];
	[self setTimer:timer];
}

- (void)startTimer
{
	NSRunLoop *runLoop = [NSRunLoop currentRunLoop];
	[runLoop addTimer:_timer forMode:NSRunLoopCommonModes];
}

- (void)reactToTimer:(NSTimer *)timer
{
	BOOL taskIsRunnung = [_task isRunning];
	if (taskIsRunnung) {
		if (_isInterrupted)
			return;
		else
			[self feedback];
	}
	else {
		if (_isInterrupted) {
			[self cleanupFileSystem];
			[self flagAbortedFileNames];
			[self terminate];
		}
		else {
			int status = [_task terminationStatus];
			if (status == 0) {
				[self updateSpeedStats];
				[self updateSpeed];
			}
			[self updateProgress];
			[self cleanupFileSystem];
			[self setInputIndex:_inputIndex + 1];
			if (_inputIndex < _inputCount)
				[self processNextFile];
			else
				[self terminate];
		}
	}
}

- (void)updateSpeedStats
{
	uint64_t now = nanosecondsSinceBoot();
	InputFile *file = [_inputFiles objectAtIndex:_inputIndex];
	_speedStats.bytes += [file size];
	_speedStats.nanoseconds += (now - _taskStartTime);
}

- (void)updateSpeed
{
	dataFormat_t *dataFormat = [_options dataFormat];
	double theSpeed = cast(double)_speedStats.bytes / _speedStats.nanoseconds;
	dataFormat->conversionSpeed = theSpeed;
	[self setSpeed:theSpeed];
}

- (void)updateProgress
{
	InputFile *file = [_inputFiles objectAtIndex:_inputIndex];
	uint64_t progress = _definitiveProgress + [file size];
	[self setDefinitiveProgress:progress];
	[_feedbacker feedbackProgress:_definitiveProgress];
}

- (void)cleanupFileSystem
{
	NSFileManager *fileManager = [NSFileManager defaultManager];
	NSError *error = nil;
	int status = [_task terminationStatus];
	if (status == 0) {
		NSString *finalPath = [self finalOutputPath];
		[fileManager moveItemAtPath:_outputPath toPath:finalPath error:&error];
	}
	else {
		[fileManager removeItemAtPath:_outputPath error:&error];
		[self flagTheFileAtIndex:_inputIndex];
	}
}

- (void)terminate
{
	[_timer invalidate];
	[self writeSpeed];
	if ([_faultyInputNames count] > 0)
		[_feedbacker feedbackFaultyInputNames:_faultyInputNames];
	else
		[_feedbacker feedbackSuccess];
}

- (void)flagAbortedFileNames
{
	for (unsigned long i = _inputIndex + 1; i < _inputCount; i++)
		[self flagTheFileAtIndex:i];
}

- (void)flagTheFileAtIndex:(unsigned long)index
{
	InputFile *file = [_inputFiles objectAtIndex:index];
	NSString *inputPath = [file path];
	[_faultyInputNames addObject:[inputPath lastPathComponent]];
}

- (void)feedback
{
	uint64_t now = nanosecondsSinceBoot();
	double estimatedPartialProgress = progressEstimation(now - _taskStartTime, _speed, _inputSize);
	estimatedPartialProgress = MIN(estimatedPartialProgress, _inputSize - 0.5);
	double estimatedTotalProgress = _definitiveProgress + estimatedPartialProgress;
	estimatedTotalProgress = MAX(estimatedTotalProgress, 0.5);
	[_feedbacker feedbackProgress:estimatedTotalProgress];
	double estimatedRemainingBytes = _totalInputSize - estimatedTotalProgress;
	double estimatedRemainingTime = (estimatedRemainingBytes / estimatedTotalProgress) * (now - _startTime);
	[_feedbacker feedbackRemainingTime:estimatedRemainingTime];
}

- (void)processNextFile
{
	InputFile *file = [_inputFiles objectAtIndex:_inputIndex];
	[self initTaskStartTime];
	NSMutableArray *args = [_commonArgs mutableCopy];
	NSString *inputPath = [file path];
	uint64_t size = [file size];
	[self setInputSize:size];
	[_feedbacker feedbackCurrentFileName:[inputPath lastPathComponent]];
	[args addObject:inputPath];
	[self initOutputPath];
	[args addObject:_outputPath];
	[self initTask];
	[_task setArguments:args];
	[_task launch];
}

- (void)initTaskStartTime
{
	uint64_t now = nanosecondsSinceBoot();
	[self setTaskStartTime:now];
	if (_inputIndex == 0)
		[self setStartTime:now];
}

- (void)initOutputPath
{
	NSFileManager *fileManager = [NSFileManager defaultManager];
	NSString *candidate = nil;
	do {
		unsigned r = cast(unsigned)random();
		NSString *outputName = [NSString stringWithFormat:@".AudioConverter_%x", r];
		candidate = [_outputDirectoryPath stringByAppendingPathComponent:outputName];
	} while ([fileManager fileExistsAtPath:candidate]);
	[self setOutputPath:candidate];
}

- (void)initTask
{
	NSFileHandle *devNullFileHandle = [NSFileHandle fileHandleWithNullDevice];
	NSTask *task = [NSTask new];
	[task setCurrentDirectoryPath:_outputDirectoryPath];
	[task setLaunchPath:@"/usr/bin/afconvert"];
	[task setStandardInput:devNullFileHandle];
	[task setStandardOutput:devNullFileHandle];
	[task setStandardError:devNullFileHandle];
	[self setTask:task];
}

- (NSString *)finalOutputPath
{
	fileFormat_t *fileFormat = [_options fileFormat];
	NSString *extension = fileFormat->extension;
	NSFileManager *fileManager = [NSFileManager defaultManager];
	InputFile *inputFile = [_inputFiles objectAtIndex:_inputIndex];
	NSString *inputPath = [inputFile path];
	NSString *inputName = [inputPath lastPathComponent];
	NSString *nameBody = [inputName stringByDeletingPathExtension];
	NSString *candidate = nil;
	unsigned i = 0;
	do {
		if (i > 0)
			nameBody = [nameBody stringByAppendingFormat:@"-%u", i];
		NSString *outputName = [nameBody stringByAppendingPathExtension:extension];
		candidate = [_outputDirectoryPath stringByAppendingPathComponent:outputName];
		i++;
	} while ([fileManager fileExistsAtPath:candidate]);
	return candidate;
}

// bytes per nanosecond = (decimal) gigabytes per second
#define minSpeed 0.00005

- (void)readSpeed
{
	dataFormat_t *dataFormat = [_options dataFormat];
	NSUserDefaults *defaults = [NSUserDefaults standardUserDefaults];
	NSString *key = [@"speed_v1_" stringByAppendingString:dataFormat->code];
	double speed = [defaults doubleForKey:key];
	if (speed >= minSpeed)
		dataFormat->conversionSpeed = speed;
}

- (void)writeSpeed
{
	dataFormat_t *dataFormat = [_options dataFormat];
	NSUserDefaults *defaults = [NSUserDefaults standardUserDefaults];
	double speed = dataFormat->conversionSpeed;
	if (speed >= minSpeed) {
		NSString *key = [@"speed_v1_" stringByAppendingString:dataFormat->code];
		[defaults setDouble:speed forKey:key];
	}
}

@end
