//
//  MainController.m
//

#import "MainController.h"
#import "Converter.h"
#import "DataFormat.h"
#import "Functions.h"

typedef struct {
	NSSize defaultSize;
	NSSize minSize;
	NSSize maxSize;
} viewSizes_t;

@interface MainController ()

@property (strong) NSOpenPanel *openPanel;
@property (strong) NSOpenPanel *savePanel;
@property (strong) Converter *converter;

// override attributes
@property (readwrite) BOOL isInterrupted;
@property (readwrite) BOOL usesVariableBitRate;

- (void)initOpenPanel;
- (void)initSavePanel;
- (void)convertWithOptions:(ConversionOptions *)options;
- (void)initConverterWithOptions:(ConversionOptions *)options;
- (void)takeContentView:(NSView *)view withSizes:(viewSizes_t *)sizes;
- (NSRect)windowFrameForContentSize:(NSSize)size;
- (void)makeMainWindowClosable:(BOOL)shouldBeClosable;

@end

@implementation MainController

//--------------------------------------
#pragma mark property implementations

synth(mainWindow);

synth(optionsView);
synth(progressView);
synth(successView);
synth(errorView);

synth(optionsController);
synth(feedbackController);

synth(openPanel);
synth(savePanel);
synth(converter);

synth(isInterrupted);
synth(usesVariableBitRate);

//--------------------------------------
#pragma mark overridden methods

- (void)awakeFromNib
{
	[self initOpenPanel];
	[self initSavePanel];
	[self showOptionsView];
}

//--------------------------------------
#pragma mark protocol methods

// NSWindowDelegate protocol method
- (void)windowWillClose:(NSNotification *)notification
{
	NSWindow *window = cast(NSWindow *)[notification object];
	if (window == _mainWindow && [_mainWindow contentView] == _optionsView)
		[NSApp terminate:self];
}

//--------------------------------------
#pragma mark public methods

- (void)showOptionsView
{
	viewSizes_t optionsViewSizes = { { 576, 420 }, { 576, 420 }, { 616, 420 } };
	[self takeContentView:_optionsView withSizes:&optionsViewSizes];
}

- (void)showProgressView
{
	viewSizes_t progressViewSizes = { { 576, 138 }, { 576, 138 }, { 616, 138 } };
	[self takeContentView:_progressView withSizes:&progressViewSizes];
}

- (void)showSuccessView
{
	viewSizes_t successViewSizes = { { 576, 138 }, { 576, 138 }, { 616, 138 } };
	[self takeContentView:_successView withSizes:&successViewSizes];
}

- (void)showErrorView
{
	viewSizes_t errorViewSizes = { { 576, 233 }, { 576, 233 }, { 616, 420 } };
	[self takeContentView:_errorView withSizes:&errorViewSizes];
}

- (void)proceedWithOptions:(ConversionOptions *)options
{
	long openPanelResult = [_openPanel runModal];
	if (openPanelResult == NSOKButton) {
		long savePanelResult = [_savePanel runModal];
		if (savePanelResult == NSOKButton)
			[self convertWithOptions:options];
	}
}

- (void)stop
{
	[self setIsInterrupted:YES];
	[_converter stop];
}

//--------------------------------------
#pragma mark private methods

#define numberOfFileExtensions 22

static NSString *fileExtensions[numberOfFileExtensions] = {
	@"3g2", @"3gp", @"aac", @"ac3", @"adts", @"aif", @"aifc", @"aiff", @"amr", @"au",
	@"caf", @"m4a", @"m4r", @"mp1", @"mp2", @"mp3", @"mp4", @"mpa", @"mpeg", @"sd2",
	@"snd", @"wav"
};

- (void)initOpenPanel
{
	NSOpenPanel *panel = [NSOpenPanel openPanel];
	[panel setMessage:@"► 1. Please choose the files to convert:"];
	unsigned sysVersion = systemVersion();
	NSString *prompt = (sysVersion >= 0x0a0700) ? @"Convert Files" : @" Convert Files ";
	[panel setPrompt:prompt];
	[panel setAllowsMultipleSelection:YES];
	[panel setResolvesAliases:YES];
	NSArray *fileTypes = [NSArray arrayWithObjects:fileExtensions count:numberOfFileExtensions];
	[panel setAllowedFileTypes:fileTypes];
	[self setOpenPanel:panel];
}

- (void)initSavePanel
{
	NSOpenPanel *panel = [NSOpenPanel openPanel];
	[panel setMessage:@"► 2. Please choose the destination folder:"];
	unsigned sysVersion = systemVersion();
	NSString *prompt = (sysVersion >= 0x0a0700) ? @"Save Here" : @" Save Here ";
	[panel setPrompt:prompt];
	[panel setTitle:@"Save"];
	[panel setCanChooseFiles:NO];
	[panel setCanChooseDirectories:YES];
	[panel setCanCreateDirectories:YES];
	[self setSavePanel:panel];
}

- (void)convertWithOptions:(ConversionOptions *)options
{
	dataFormat_t *dataFormat = [options dataFormat];
	[self setUsesVariableBitRate:dataFormat->variableBitRate];
	[self setIsInterrupted:NO];
	[self initConverterWithOptions:options];
	[_feedbackController resetProgressView];
	[self showProgressView];
	[_converter start];
}

- (void)initConverterWithOptions:(ConversionOptions *)options
{
	Converter *converter = [Converter new];
	[converter setOptions:options];
	[converter setInputURLs:[_openPanel URLs]];
	[converter setOutputDirectoryURL:[[_savePanel URLs] objectAtIndex:0]];
	[converter setFeedbacker:_feedbackController];
	[self setConverter:converter];
}

- (void)takeContentView:(NSView *)view withSizes:(viewSizes_t *)sizes
{
	NSRect frame = [self windowFrameForContentSize:sizes->defaultSize];
	[[_mainWindow contentView] setHidden:YES];
	[_mainWindow setFrame:frame display:YES animate:YES];
	[view setHidden:NO];
	[view setFrameOrigin:NSZeroPoint];
	[view setFrameSize:sizes->defaultSize];
	[_mainWindow setContentView:view];
	[_mainWindow setContentMinSize:sizes->minSize];
	[_mainWindow setContentMaxSize:sizes->maxSize];
	[self makeMainWindowClosable:(view == _optionsView)];
}

- (NSRect)windowFrameForContentSize:(NSSize)size
{
	NSSize oldContentSize = [[_mainWindow contentView] frame].size;
	CGFloat widthDifference = size.width - oldContentSize.width;
	CGFloat heightDifference = size.height - oldContentSize.height;
	NSRect newFrame = [_mainWindow frame];
	newFrame.origin.y -= heightDifference;
	newFrame.size.width += widthDifference;
	newFrame.size.height += heightDifference;
	return newFrame;
}

- (void)makeMainWindowClosable:(BOOL)shouldBeClosable
{
	unsigned long styleMask = [_mainWindow styleMask];
	if (shouldBeClosable)
		styleMask |= NSClosableWindowMask;
	else
		styleMask &= ~cast(unsigned long)NSClosableWindowMask;
	[_mainWindow setStyleMask:styleMask];
}

@end
