//
//  OptionsController.m
//

#import "OptionsController.h"
#import "ChannelLayout.h"
#import "ConversionOptions.h"
#import "DataFormat.h"
#import "FileFormat.h"
#import "MainController.h"

@interface OptionsController ()

@property (strong) NSMutableArray *dataFormatMenus;

- (void)initDataFormatMenus;

- (void)initLayoutPopup;
- (void)initFileFormatPopup;
- (void)initSampleRateCombo;
- (void)initBitRateCombo;
- (void)initStrategyPopup;

- (void)resetUI;

- (void)updateDataFormatPopup;
- (void)updateLayoutPopup;
- (void)updateChennelsTextField;
- (void)updateBitRateCombo;
- (void)updateStrategyPopup;
- (void)updateCodecQualitySlider;
- (void)updateFillerCheckbox;

- (void)updateSrcQualityTextField;
- (void)updateSrcComplexityTextField;
- (void)updateCodecQualityTextField;

- (ConversionOptions *)readOptions;

- (dataFormat_t *)currentDataFormat;
- (NSMenu *)newDataFormatMenuForFileFormat:(fileFormat_t *)fileFormat;

@end

@implementation OptionsController

//--------------------------------------
#pragma mark property implementations

synth(layoutPopup);
synth(channelsTextField);
synth(fileFormatPopup);
synth(dataFormatPopup);
synth(sampleRateCombo);
synth(srcQualitySlider);
synth(srcComplexitySlider);
synth(bitRateCombo);
synth(strategyPopup);
synth(codecQualitySlider);
synth(fillerCheckbox);
synth(srcQualityTextField);
synth(srcComplexityTextField);
synth(codecQualityTextField);
synth(formatter);
synth(mainController);

synth(dataFormatMenus);

//--------------------------------------
#pragma mark overridden methods

- (void)awakeFromNib
{
	[self initDataFormatMenus];
    [self initLayoutPopup];
	[self initFileFormatPopup];
	[self initSampleRateCombo];
	[self initBitRateCombo];
	[self initStrategyPopup];
	[self resetUI];
}

//--------------------------------------
#pragma mark public methods

- (IBAction)getLayoutFrom:(NSPopUpButton *)sender
{
	[self updateChennelsTextField];
}

- (IBAction)getChannelsFrom:(NSTextField *)sender
{
}

- (IBAction)getFileFormatFrom:(NSPopUpButton *)sender
{
	[self updateDataFormatPopup];
}

- (IBAction)getDataFormatFrom:(NSPopUpButton *)sender
{
	[self updateLayoutPopup];
	[self updateBitRateCombo];
	[self updateStrategyPopup];
	[self updateCodecQualitySlider];
	[self updateFillerCheckbox];
}

- (IBAction)getSampleRateFrom:(NSComboBox *)sender
{
}

- (IBAction)getSrcQualityFrom:(NSSlider *)sender
{
	[self updateSrcQualityTextField];
}

- (IBAction)getSrcComplexityFrom:(NSSlider *)sender
{
	[self updateSrcComplexityTextField];
}

- (IBAction)getBitRateFrom:(NSComboBox *)sender
{
}

- (IBAction)getStrategyFrom:(NSPopUpButton *)sender
{
}

- (IBAction)getCodecQualityFrom:(NSSlider *)sender
{
	[self updateCodecQualityTextField];
}

- (IBAction)getFillerFlagFrom:(NSButton *)sender
{
}

- (IBAction)proceed:(id)sender
{
	ConversionOptions *options = [self readOptions];
	[_mainController proceedWithOptions:options];
}

- (IBAction)quit:(id)sender
{
    [NSApp terminate:self];
}

//--------------------------------------
#pragma mark private methods

- (void)initDataFormatMenus
{
	NSMutableArray *menus = [NSMutableArray arrayWithCapacity:numberOfFileFormats];
	for (unsigned i = 0; i < numberOfFileFormats; i++)
		[menus addObject:[self newDataFormatMenuForFileFormat:(fileFormats + i)]];
	[self setDataFormatMenus:menus];
}

- (void)initLayoutPopup
{
	NSMenu *menu = [_layoutPopup menu];
	for (unsigned i = 0; i < numberOfChannelLayouts; i++) {
		channelLayout_t *layout = channelLayouts + i;
		NSMenuItem *item = nil;
		if (layout->name == nil)
			item = [NSMenuItem separatorItem];
		else {
			item = [NSMenuItem new];
			[item setTitle:layout->name];
			if (layout->tooltip != nil)
				[item setToolTip:layout->tooltip];
		}
		[menu addItem:item];
	}
}

- (void)initFileFormatPopup
{
	for (unsigned i = 0; i < numberOfFileFormats; i++)
		[_fileFormatPopup addItemWithTitle:fileFormats[i].name];
}

- (void)initSampleRateCombo
{
	NSString *sampleRateStrings[] = { @"Original", @"22050", @"44100", @"48000", @"96000" };
	for (unsigned i = 0; i < 5; i++)
		[_sampleRateCombo addItemWithObjectValue:sampleRateStrings[i]];
}

- (void)initBitRateCombo
{
	NSString *bitRateStrings[] = { @"128", @"160", @"192", @"256", @"320" };
	for (unsigned i = 0; i < 5; i++)
		[_bitRateCombo addItemWithObjectValue:bitRateStrings[i]];
}

- (void)initStrategyPopup
{
	NSString *strategyStrings[] = { @"Constant bit rate", @"Average bit rate", @"VBR but constrained", @"Variable bit rate" };
	for (unsigned i = 0; i < 4; i++)
		[_strategyPopup addItemWithTitle:strategyStrings[i]];
}

#define defaultFileFormatIndex 2
#define defaultSampleRateIndex 0
#define defaultSrcQuality 4
#define defaultSrcComplexity 2
#define defaultCodecQuality 4

- (void)resetUI
{
	[_fileFormatPopup selectItemAtIndex:defaultFileFormatIndex];
	[_sampleRateCombo selectItemAtIndex:defaultSampleRateIndex];
	[_srcQualitySlider setIntValue:defaultSrcQuality];
	[_srcComplexitySlider setIntValue:defaultSrcComplexity];
	[_codecQualitySlider setIntValue:defaultCodecQuality];
	[self getFileFormatFrom:_fileFormatPopup];
	[self getSampleRateFrom:_sampleRateCombo];
	[self getSrcQualityFrom:_srcQualitySlider];
	[self getSrcComplexityFrom:_srcComplexitySlider];
}

- (void)updateDataFormatPopup
{
	unsigned fileFormatIndex = cast(unsigned)[_fileFormatPopup indexOfSelectedItem];
	[_dataFormatPopup setMenu:[_dataFormatMenus objectAtIndex:fileFormatIndex]];
	[_dataFormatPopup selectItemAtIndex:fileFormats[fileFormatIndex].indexOfDefaultDataFormat];
	[self getDataFormatFrom:_dataFormatPopup];
}

#define channelLayout_original 0
#define channelLayout_discreteInOrder 2

- (void)updateLayoutPopup
{
	dataFormat_t *dataFormat = [self currentDataFormat];
	BOOL shouldEnable = dataFormat->variableLayout;
	BOOL isEnabled = [_layoutPopup isEnabled];
	if (isEnabled && !shouldEnable)
		[_layoutPopup selectItemAtIndex:channelLayout_original];
	[_layoutPopup setEnabled:shouldEnable];
	[self getLayoutFrom:_layoutPopup];
}

#define defaultChannelCount 2

- (void)updateChennelsTextField
{
	unsigned layoutIndex = cast(unsigned)[_layoutPopup indexOfSelectedItem];
	BOOL shouldEnable = (layoutIndex == channelLayout_discreteInOrder);
	BOOL isEnabled = [_channelsTextField isEnabled];
	if (isEnabled && !shouldEnable)
		[_channelsTextField setFormatter:nil];
	else if (!isEnabled && shouldEnable) {
		[_channelsTextField setFormatter:_formatter];
		[_channelsTextField setIntValue:defaultChannelCount];
	}
	if (layoutIndex == channelLayout_original)
		[_channelsTextField setStringValue:@"Automatic"];
	else if (layoutIndex > channelLayout_discreteInOrder)
		[_channelsTextField setIntValue:(channelLayouts[layoutIndex].tag & 0x0FFFF)];
	[_channelsTextField setEnabled:shouldEnable];
}

#define defaultBitRateIndex 2

- (void)updateBitRateCombo
{
	dataFormat_t *dataFormat = [self currentDataFormat];
	BOOL shouldEnable = dataFormat->variableBitRate;
	BOOL isEnabled = [_bitRateCombo isEnabled];
	if (isEnabled && !shouldEnable)
		[_bitRateCombo setStringValue:@"Automatic"];
	else if (!isEnabled && shouldEnable)
		[_bitRateCombo selectItemAtIndex:defaultBitRateIndex];
	[_bitRateCombo setEnabled:shouldEnable];
}

#define defaultStrategyIndex_forDisabled 0
#define defaultStrategyIndex_forEnabled 3

- (void)updateStrategyPopup
{
	dataFormat_t *dataFormat = [self currentDataFormat];
	BOOL shouldEnable = dataFormat->variableBitRate;
	BOOL isEnabled = [_strategyPopup isEnabled];
	if (isEnabled && !shouldEnable)
		[_strategyPopup selectItemAtIndex:defaultStrategyIndex_forDisabled];
	else if (!isEnabled && shouldEnable)
		[_strategyPopup selectItemAtIndex:defaultStrategyIndex_forEnabled];
	[_strategyPopup setEnabled:shouldEnable];
}

- (void)updateCodecQualitySlider
{
	dataFormat_t *dataFormat = [self currentDataFormat];
	BOOL shouldEnable = dataFormat->variableQuality;
	BOOL isEnabled = [_codecQualitySlider isEnabled];
	if (isEnabled && !shouldEnable)
		[_codecQualitySlider setIntValue:defaultCodecQuality];
	[_codecQualitySlider setEnabled:shouldEnable];
	[self getCodecQualityFrom:_codecQualitySlider];
}

- (void)updateFillerCheckbox
{
	dataFormat_t *dataFormat = [self currentDataFormat];
	BOOL shouldEnable = dataFormat->worksWithFillerOption;
	BOOL isEnabled = [_fillerCheckbox isEnabled];
	if (isEnabled && !shouldEnable)
		[_fillerCheckbox setIntValue:NO];
	else if (!isEnabled && shouldEnable)
		[_fillerCheckbox setIntValue:YES];
	[_fillerCheckbox setEnabled:shouldEnable];
}

static NSString *qualityStrings[] = { @"Faster", @"Fast", @"Normal", @"Better", @"Best" };

- (void)updateSrcQualityTextField
{
	unsigned srcQuality = cast(unsigned)[_srcQualitySlider intValue];
	[_srcQualityTextField setStringValue:qualityStrings[srcQuality]];
}

- (void)updateSrcComplexityTextField
{
	NSString *complexityStrings[] = { @"Linear", @"Normal", @"Mastering" };
	unsigned srcComplexity = cast(unsigned)[_srcComplexitySlider intValue];
	[_srcComplexityTextField setStringValue:complexityStrings[srcComplexity]];
}

- (ConversionOptions *)readOptions
{
	ConversionOptions *options = [ConversionOptions new];
	[options setChannelLayout:channelLayouts + [_layoutPopup indexOfSelectedItem]];
	[options setNumberOfChannels:cast(unsigned)[_channelsTextField intValue]];
	[options setFileFormat:fileFormats + [_fileFormatPopup indexOfSelectedItem]];
	[options setDataFormat:[self currentDataFormat]];
	[options setSampleRate:cast(unsigned)[_sampleRateCombo intValue]];
	[options setSrcQuality:cast(unsigned)[_srcQualitySlider intValue]];
	[options setSrcComplexity:cast(unsigned)[_srcComplexitySlider intValue]];
	[options setBitRate:cast(unsigned)[_bitRateCombo intValue]];
	[options setStrategy:cast(unsigned)[_strategyPopup indexOfSelectedItem]];
	[options setCodecQuality:cast(unsigned)[_codecQualitySlider intValue]];
	[options setFillerFlag:cast(BOOL)[_fillerCheckbox intValue]];
	return options;
}

- (void)updateCodecQualityTextField
{
	unsigned codecQuality = cast(unsigned)[_codecQualitySlider intValue];
	[_codecQualityTextField setStringValue:qualityStrings[codecQuality]];
	BOOL shouldBeBlack = [_codecQualitySlider isEnabled];
	BOOL isBlack = ([_codecQualityTextField tag] == 0);
	if (isBlack && !shouldBeBlack) {
		[_codecQualityTextField setTextColor:[NSColor disabledControlTextColor]];
		[_codecQualityTextField setTag:1];
	}
	else if (!isBlack && shouldBeBlack) {
		[_codecQualityTextField setTextColor:[NSColor controlTextColor]];
		[_codecQualityTextField setTag:0];
	}
}

- (dataFormat_t *)currentDataFormat
{
	fileFormat_t *fileFormat = fileFormats + [_fileFormatPopup indexOfSelectedItem];
	unsigned index = cast(unsigned)[_dataFormatPopup indexOfSelectedItem];
	return (fileFormat->dataFormats)[index];
}

- (NSMenu *)newDataFormatMenuForFileFormat:(fileFormat_t *)fileFormat
{
	NSMenu *menu = [NSMenu new];
	for (dataFormat_t **handle = fileFormat->dataFormats; *handle != NULL; handle++) {
		NSMenuItem *item = nil;
		dataFormat_t *pointer = *handle;
		if (pointer == SEPERATOR)
			item = [NSMenuItem separatorItem];
		else {
			item = [NSMenuItem new];
			[item setTitle:pointer->name];
		}
		[menu addItem:item];
	}
	return menu;
}

@end
